package GrabEnv;

use strict;

use vars qw( @ISA @EXPORT @EXPORT_OK $VERSION );

use Carp;
use IO::File;
use File::Spec;

use POSIX;

@ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

@EXPORT = ( );
@EXPORT_OK = qw( grabenv grabenv_File grabenv_Expect diffenv envstr );

$VERSION = '0.08';


# Preloaded methods go here.

# we redefine the shells' prompt to this so we know what to look for
my $Prompt = 'GrabEnvPrompt> ';

# a compendium of shells and how to make them stop sending extraneous
# editing stuff and how to set their prompts
my %shells = (
	     sh => { inter => 'i',
		     nostartup => 'p',
		     verbose => 'v',
		     echo => 'x',
		     prompt => "PS1='$Prompt'" },

	    ksh => { inter => 'i',
		     nostartup => 'p',
		     verbose => 'v',
		     echo => 'x',
		     prompt => "set +o emacs +o vi +o gmacs; PS1='$Prompt'"
		   },

	    csh => { inter => 'i',
		     nostartup => 'f',
		     echo => 'x',
		     verbose => 'v',
		     prompt => "set prompt = '$Prompt'"
		   },

	   tcsh => { inter => 'i',
		     nostartup => 'f',
		     echo => 'x',
		     verbose => 'v',
		     prompt => "unset edit; unset cwdcmd; set prompt = '$Prompt'"
		   },
	  );


our %Opts = ( Startup => 0, 
	      Echo => 0, 
	      Verbose => 0,
	      STDERR => undef,
	      STDOUT => undef );

my %lcOpts = map { lc $_ => $_ } keys %Opts;

# default to using the File mode.
*grabenv = *grabenv_File;


# use temporary script files and output files to get the environment
# requires that a shell have a '-i' flag to act as an interactive shell
sub grabenv_File
{
  my $uopt = pop @_ if 'HASH' eq ref $_[-1];
  my ( $shell, $cmd ) = @_;

  my %opts = chk_opts( $uopt );

  croak( __PACKAGE__, ": unknown shell" ) unless defined $shells{$shell};

  # create script to run
  my $script = POSIX::tmpnam();
  my $s_fh = new IO::File $script, 'w'
    or croak( __PACKAGE__, ": unable to open tmp script file `$script'" );

  # file to contain environment; we create it to ensure that
  # it's ours
  my $env_out = POSIX::tmpnam();
  my $e_fh = new IO::File $env_out, O_CREAT | O_EXCL
    or croak( __PACKAGE__, ": unable to create tmp env file `$env_out'" );
  $e_fh->close();

  print $s_fh $cmd, "\n";
  print $s_fh qq{perl -MData::Dumper -e 'open F,">$env_out"||exit(1);print F Data::Dumper->Dump([\\%ENV], [\"NewEnv\"])'}, "\n";
  $s_fh->close();

  # now, invoke shell

  # redirect STDERR & STDOUT to either /dev/null or somewhere the user points
  # us to.  we can use any old filehandle here, as we're protected
  # by being inside a package.

  my $stdout = $opts{STDOUT} || File::Spec->devnull();

  my $stderr = $opts{STDERR} || File::Spec->devnull();

  open( SAVEOUT, ">&STDOUT" ) 
    or croak( __PACKAGE__,  ': error duping STDOUT' );
  open( SAVEERR, ">&STDERR" ) 
    or croak( __PACKAGE__,  ': error duping STDERR' );
  
  open( STDERR, ">$stderr" ) or 
    croak( __PACKAGE__, ": unable to redirect STDERR to $stderr" );

  open( STDOUT, ">$stdout" ) or 
    croak( __PACKAGE__, ": unable to redirect STDOUT to $stdout" );

  my $res;
  eval { 
    select STDERR; $| = 1;
    select STDOUT; $| = 1;
    
    my $options = '-' . 
      join( '',
	    $shells{$shell}{inter},
	    ($opts{Startup} ? '' : $shells{$shell}{nostartup} ),
	    ($opts{Echo} ? $shells{$shell}{echo} : '' ),
	    ($opts{Verbose} ? $shells{$shell}{verbose} : '' ),
	  );

    $res = system( "$shell $options < $script" );
  };

  close STDOUT;
  close STDERR;

  open STDOUT, ">&SAVEOUT";
  open STDERR, ">&SAVEERR";

  close SAVEOUT;
  close SAVEERR;

  unlink $script;

  chomp( $@ ) if $@;
  croak( __PACKAGE__, ": $@" ) if $@;

  croak( __PACKAGE__, ": error running script\n" ) unless $res == 0;

  my $fh = new IO::File $env_out, 'r'
    or croak( __PACKAGE__, ": unable to open tmp env file `$env_out'" );

  # suck in the entire file as one line.
  {
    local $/;
    $res = <$fh>;
  }

  $fh->close();
  unlink $env_out;

  my $NewEnv;
  eval { eval $res };
  croak $@ if $@;

  %$NewEnv;
}

# interactively query a running shell via Expect to get environment
use Expect;
sub grabenv_Expect
{
  my $uopt = pop @_ if 'HASH' eq ref $_[-1];
  my ( $shell, $cmd, $timeout ) = @_;

  my %opts = chk_opts( $uopt );

  croak( "unknown shell" ) unless defined $shells{$shell};

  # if no timeout is specified, make it 1000 seconds.
  # this does not give the caller the option of an unlimited timeout.
  $timeout ||= 1000;

  my $options = '-' . 
    join('',
	 $shells{$shell}{inter},
	 ($opts{Startup} ? '' : $shells{$shell}{nostartup}),
	 ($opts{Verbose} ? $shells{$shell}{verbose} : '' ),
	 ($opts{Echo} ? $shells{$shell}{echo} : '' ),
	);

  my $c = Expect->spawn( $shell, $options );
  $c->log_stdout(0);

  # grab something, doesn't matter what.  need to get that first prompt
  $c->expect( $timeout, -re => '.+' ) 
    or croak( "error getting initial prompt: ", $c->exp_error );

  # change prompt and wait for it to return, to get in sync
  $c->print( $shells{$shell}{prompt}, "\n" );

  # grab actual prompt; need CRLF in front to avoid getting the one
  # from the previous command
  $c->expect( $timeout, "\r\n$Prompt" ) 
    or croak( "couldn't get revised prompt: ", $c->exp_error );

  # send out the command to change the environment and wait for the
  # prompt
  $c->print( "$cmd\n" );
  $c->expect( $timeout, $Prompt ) or
    croak( "error after sending command: ", $c->exp_error);

  # use perl to get the environment.  the printenv command isn't
  # always available
  my $getenv = 
   "perl -MData::Dumper -e 'print Data::Dumper->Dump([\\%ENV], [\"NewEnv\"])'";
  $c->print( "$getenv\n" );
  $c->expect( $timeout, "\r\n$Prompt" ) 
    or croak( "error getting environment: ", $c->exp_error);

  my $res = $c->exp_before;

  # kill up to Data::Dumper output, as the result contains the getenv
  # command.  unfortunately, we can't turn echoing off using exp_stty;
  # it doesn't seem to work on solaris
  $res =~  s/.*^\$NewEnv/\$NewEnv/ms;

  my $NewEnv;
  eval { eval $res };
  croak $@ if $@;

  %$NewEnv;
}

sub chk_opts
{
  my $opt = shift || {};
  
  local $Carp::CarpLevel = $Carp::CarpLevel - 1;

  my @notvalid = grep { ! defined $lcOpts{ lc $_ } } keys %$opt;
  croak( __PACKAGE__, "::grabenv illegal option(s): @notvalid\n" )
    if @notvalid;

  my %opts = ( %Opts, map { $lcOpts{lc $_} => $opt->{$_} } keys %$opt );
  %opts;
}

sub diffenv
{
  my ( $new, $old ) = @_;

  $old ||= \%ENV;

  my %diffs;
  while( my ( $key, $val ) = each %$new )
  {
    $diffs{$key} = $val
      unless exists $old->{$key} && $old->{$key} eq $new->{$key};
  }

  %diffs;
}

# construct argument for "env" command
sub envstr
{
  my ( $new, $attr ) = @_;

  my %attr = ( ZapDeleted => 1, 
	       DiffsOnly => 1, 
	       Ignore => [],
	       $attr ? %$attr : () );

  my @env;

  my %ignore = ( map { $_ => 1 } @{$attr{Ignore}} );

  if ( $attr{DiffsOnly} )
  {
    while( my ($key, $value) = each %$new )
    {
      next if exists $ignore{$key};

      if ( !defined $ENV{$key} || $ENV{$key} ne $value )
      {
	push @env, "$key=" . _shell_escape($value);
      }
    }
  }

  else
  {
    while( my ($key, $value) = each %$new )
    {
      next if exists $ignore{$key};
	push @env, "$key=" . _shell_escape($value);
    }
  }

  if ( $attr{ZapDeleted} )
  {
    foreach ( keys %ENV )
    {
      next if exists $ignore{$_};
      push @env, "$_=" unless defined $new->{$_};
    }
  }

  return join( ' ', @env);
}

my $MAGIC_CHARS = q/\\\$"'!*{};()[]/; #";
$MAGIC_CHARS =~ s/(\W)/\\$1/g;

sub _shell_escape
{
  my $str = shift;


  # if there's white space, single quote the entire word.  however,
  # since single quotes can't be escaped inside single quotes,
  # isolate them from the single quoted part and escape them.
  # i.e., the string a 'b turns into 'a '\''b' 
  if ( $str =~ /\s/ )
  {
    # isolate the lone single quotes
    $str =~ s/'/'\\''/g;

    # quote the whole string
    $str = "'$str'";

    # remove obvious duplicate quotes.
    $str =~ s/(^|[^\\])''/$1/g;
  }
  elsif ( $str =~ /[$MAGIC_CHARS]/ )
  {
    $str =~  s/([$MAGIC_CHARS])/\\$1/g;
  }

  # empty string
  elsif ( $str eq '' )
  {
    $str = "''";
  }
  $str;
}



# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

GrabEnv - make a snapshot of the environment after running a command

=head1 SYNOPSIS

  use GrabEnv qw( grabenv grabenv_File grabenv_Expect diffenv envstr );

  %NewENV = grabenv( $shell, $command, \%opts );
  %NewENV = grabenv( $shell, $command, $timeout, \%opts );

  %NewENV = grabenv_Expect( $shell, $command, \%opts );
  %NewENV = grabenv_Expect( $shell, $command, $timeout, \%opts );

  %NewENV = grabenv_File( $shell, $command, \%opts );

  %diffEnv = diffenv( \%NewENV );
  %diffEnv = diffenv( \%NewENV, \%OldEnv );

  $env_str = envstr( \%NewENV );

=head1 DESCRIPTION

B<GrabEnv> provides the B<grabenv> function, which will run a command
under a given shell and return the environment after the command is
run.  In order for this to work (and make any sense) the command can't
actually be an executable; it's either a builtin command, or 
a statement which sources some setup file (or an alias to do that).

Why? Well, let's say there's a software package you'd like to run
from a Perl script, but it requires its own blend of paths and
environmental variables.  Let's say that the package provides a B<csh>
script which the user can source to set up the environment. You'd like
your Perl script to run without the user having to set up that
environment, but unfortunately, can't source the B<csh> script from
within Perl.  Or, perhaps you need to work with more than one package
each with its own, possibly conflicting environment.

B<grabenv> is there for you.  It starts up a shell, runs the command,
captures the new environment, and returns a new C<%ENV> style hash
containing the environment. 

It does this using one of two methods, either using the B<Expect>
module, which makes the shell believe it's talking to the terminal, or
by creating a script for the shell to run.  The script produces an
output file, which B<grabenv> parses.  In the latter case, the shell
is invoked in interactive mode.  Unless specified otherwise,
B<GrabEnv> tries to ensure that the user's shell start up scripts are
not invoked.  However, if those scripts are invoked, the interactive
bit is important, as some user's shell start up scripts do things
differently in batch vs. interactive mode.

By default, the script uses the non-B<Expect> mode.  You can call either
mode explicitly by using the B<grabenv_Expect> or B<grabenv_File>
functions.

There shouldn't be any difference between the two modes, but it's
possible that since using the B<File> version the shell isn't
connected to a real terminal, that may cause problems.  Conversely,
with the B<Expect> mode, it must do arcane things to the shells to
avoid problems with prompts and command line editing, and this may
break for particular versions of the shells which haven't been tested
(all tests having been done on Solaris).  It also may timeout if
network mounts are required and there are severe network delays.
The B<File> mode doesn't have a timeout, so will just hang until
things clear up.

It'll B<croak()> upon error.

=head2 Functions

=over 8

=item grabenv

  %NewENV = grabenv( $shell, $cmd, \%opt );

This simply calls B<grabenv_File()>.  For example:

  %NewENV = grabenv( 'csh', 'source /soft/ciao/bin/ciao' );

Options may be passed via a hash passed as the last argument.
The following keys are recognized:

=over 8

=item Startup

If the value of this element is non-zero, the user's shell startup
files will be invoked by the shell.  It defaults to C<0>.

=item Echo

If non-zero, put the shell in echo mode. This is only of use when the
B<STDOUT> option is specified.  It defaults to C<0>.

=item Verbose

If non-zero, put the shell in verbose mode. This is only of use when
the B<STDOUT> option is specified.  It defaults to C<0>.

=item STDERR

Normally output from the shells' standard error stream is discarded.
This may be set to a file name to which the stream should be written.
At present this has only been tested with B<grabenv_File>.

=item STDOUT

Normally output from the shells' standard output stream is discarded.
This may be set to a file name to which the stream should be written.
At present this has only been tested with B<grabenv_File>.

=back


=item grabenv_Expect

  %NewENV = grabenv_Expect( $shell, $cmd );
  %NewENV = grabenv_Expect( $shell, $cmd, \%opt );
  %NewENV = grabenv_Expect( $shell, $cmd, $timeout );
  %NewENV = grabenv_Expect( $shell, $cmd, $timeout, \%opt );

The optional B<$timeout> argument is the number of seconds to wait for
a response from the shell until giving up.  It defaults to 1000
seconds, but some scripts may require more time; only you will know.
See B<grabenv> for the meaning of the other arguments.

=item grabenv_File

  %NewENV = grabenv_File( $shell, $cmd );
  %NewENV = grabenv_File( $shell, $cmd, \%opt );

See B<grabenv> for the meaning of the arguments.

=item diffenv

  %deffENV = diffenv( \%NewENV, \%OldENV );
  %diffENV = diffenv( \%NewENV );

This is a silly convenience function that'll return the parts
of C<%NewENV> which are different from C<%OldENV>.  If C<%OldENV>
isn't specified, it compares it to C<%ENV>.  It doesn't track
things in C<%OldENV> which aren't in C<%NewENV>.

=item env

  $env_str = envstr( \%NewENV );
  $env_str = envstr( \%NewENV, \%attr );

This returns a string defining the passed environment
suitable for use by the UNIX B<env> command.  It operates on the
passed hash.  The passed attribute hash modifies its behavior; available
attributes (keys) are

=over 8

=item DiffsOnly

If true, only output keys which differ from what's in C<%ENV>.  For example,

   %NewENV = %ENV;
   $NewENV{FOOFOO} = 'googoo';
   print envstr( \%NewENV, { DiffsOnly => 1 } );

results in

   FOOFOO=googoo

rather than a zillion environmental variables.
This attribute defaults to being B<on>.

=item ZapDeleted

If true, keys which are in C<%ENV>, but not in C<%NewENV> are explicitly
deleted:

   %NewENV = %ENV;
   $ENV{FOOFOO} = 'googoo';
   print envstr( \%NewENV, { DiffsOnly => 1, ZapDeleted => 1 } ), "\n";
 
results in

   FOOFOO=

This attribute defaults to being B<on>.


=item Ignore

This should be an reference to an array containing variables to be ignored.
They will not be output.

=back



=back



=head2 EXPORT

No functions are exported by default.


=head2 ARCANITIES

Talking to interactive shells is not so easy, especially as the user
may have set up editing environments for the shell.  Things may not
work correctly; please send the author any updates on recalcitrant
shells.  Here's what's done to turn off fancy editing and thus get
things to work:

=over 8

=item ksh

    set +o emacs +o vi +o gmacs

=item tcsh

    unset edit; unset cwdcmd

=back

Additionally, B<grabenv> instructs the shell it use its own prompt.

B<grabenv> uses Perl and the B<Data::Dumper> module in the running
shell to extract the environmental variables, as the B<printenv> command
is not always built in to the shell and may not be in the user's path.
It then B<eval>'s the result.


=head1 PREREQUISITES

The B<Expect> and B<Data::Dumper> modules.

=head1 LICENSE

This software is released under the GNU General Public License.  You
may find a copy at 

   http://www.fsf.org/copyleft/gpl.html


=head1 AUTHOR

Diab Jerius ( djerius@cfa.harvard.edu )

=head1 SEE ALSO

=cut
